@@ -19,6 +19,12 @@ from mongomock import filtering
 from mongomock import helpers
 from mongomock import OperationFailure
 
+try:
+    from bson import decimal128
+    decimal_support = True
+except ImportError:
+    decimal_support = False
+
 _random = random.Random()
 
 
@@ -88,7 +94,6 @@ string_operators = [
     '$substr',
     '$toLower',
     '$toUpper',
-    '$toString',
 ]
 comparison_operators = [
     '$cmp',
@@ -107,6 +112,11 @@ set_operators = [
     '$allElementsTrue',
 ]
 
+type_convertion_operators = [
+    '$toString',
+    '$toInt'
+]
+
 
 def _avg_operation(values):
     values_list = list(v for v in values if isinstance(v, numbers.Number))
@@ -143,6 +153,12 @@ class _Parser(object):
         if not isinstance(expression, dict):
             return self._parse_basic_expression(expression)
 
+        if len(expression) > 1 and any(key.startswith('$') for key in expression):
+            raise OperationFailure(
+                'an expression specification must contain exactly one field, '
+                'the name of the expression. Found %d fields in %s'
+                % (len(expression), expression))
+
         value_dict = {}
         for k, v in six.iteritems(expression):
             if k in arithmetic_operators:
@@ -163,6 +179,8 @@ class _Parser(object):
                 return self._handle_set_operator(k, v)
             if k in string_operators:
                 return self._handle_string_operator(k, v)
+            if k in type_convertion_operators:
+                return self._handle_type_convertion_operator(k, v)
             if k in boolean_operators + \
                     text_search_operators + projection_operators:
                 raise NotImplementedError(
@@ -179,6 +197,14 @@ class _Parser(object):
 
         return value_dict
 
+    def parse_many(self, values):
+        for value in values:
+            try:
+                yield self.parse(value)
+            except KeyError:
+                if self._ignore_missing_keys:
+                    yield None
+
     def _parse_basic_expression(self, expression):
         if isinstance(expression, six.string_types) and expression.startswith('$'):
             if expression.startswith('$$'):
@@ -224,7 +250,12 @@ class _Parser(object):
             return math.sqrt(self.parse(values))
         if operator == '$subtract':
             assert len(values) == 2, 'subtract must have only 2 items'
-            res = self.parse(values[0]) - self.parse(values[1])
+            value_0 = self.parse(values[0])
+            value_1 = self.parse(values[1])
+            if isinstance(value_0, datetime.datetime) and \
+                    isinstance(value_1, (six.integer_types, float)):
+                value_1 = datetime.timedelta(milliseconds=value_1)
+            res = value_0 - value_1
             if isinstance(res, datetime.timedelta):
                 return round(res.total_seconds() * 1000)
             return res
@@ -237,7 +268,7 @@ class _Parser(object):
 
     def _handle_project_operator(self, operator, values):
         if operator in _GROUPING_OPERATOR_MAP:
-            return _GROUPING_OPERATOR_MAP[operator](self.parse(val) for val in values)
+            return _GROUPING_OPERATOR_MAP[operator](self.parse_many(values))
         if operator == '$arrayElemAt':
             key, index = values
             array = self._parse_basic_expression(key)
@@ -276,7 +307,7 @@ class _Parser(object):
             parsed = self.parse(values)
             return str(parsed).upper() if parsed is not None else ''
         if operator == '$concat':
-            parsed_list = [self.parse(value) for value in values]
+            parsed_list = list(self.parse_many(values))
             return None if None in parsed_list else ''.join([str(x) for x in parsed_list])
         if operator == '$substr':
             if len(values) != 3:
@@ -300,9 +331,6 @@ class _Parser(object):
                 raise OperationFailure('strcasecmp must have 2 items')
             a, b = str(self.parse(values[0])), str(self.parse(values[1]))
             return 0 if a == b else -1 if a < b else 1
-        if operator == '$toString':
-            parsed = self.parse(values)
-            return str(parsed) if parsed is not None else None
         # This should never happen: it is only a safe fallback if something went wrong.
         raise NotImplementedError(  # pragma: no cover
             "Although '%s' is a valid string operator for the aggregation "
@@ -407,6 +435,28 @@ class _Parser(object):
             'aggregation pipeline, it is currently not implemented '
             'in Mongomock.' % operator)
 
+    def _handle_type_convertion_operator(self, operator, values):
+        if operator == '$toString':
+            try:
+                parsed = self.parse(values)
+            except KeyError:
+                return None
+            if isinstance(parsed, bool):
+                return str(parsed).lower()
+            if isinstance(parsed, datetime.datetime):
+                return parsed.isoformat()[:-3] + 'Z'
+            return str(parsed)
+
+        if operator == '$toInt':
+            parsed = self.parse(values)
+            if decimal_support:
+                if isinstance(parsed, decimal128.Decimal128):
+                    return int(parsed.to_decimal())
+                return int(parsed)
+            raise NotImplementedError(
+                'You need to import the pymongo library to support decimal128 type.'
+            )
+
     def _handle_conditional_operator(self, operator, values):
         if operator == '$ifNull':
             field, fallback = values
@@ -922,6 +972,27 @@ def _handle_project_stage(in_collection, unused_database, options):
     return out_collection
 
 
+def _handle_add_fields_stage(in_collection, unused_database, options):
+    if not options:
+        raise OperationFailure(
+            'Invalid $addFields :: caused by :: specification must have at least one field')
+    out_collection = [dict(doc) for doc in in_collection]
+    for field, value in six.iteritems(options):
+        for in_doc, out_doc in zip(in_collection, out_collection):
+            try:
+                out_value = _parse_expression(value, in_doc, ignore_missing_keys=True)
+            except KeyError:
+                continue
+            parts = field.split('.')
+            for subfield in parts[:-1]:
+                out_doc[subfield] = out_doc.get(subfield, {})
+                if not isinstance(out_doc[subfield], dict):
+                    out_doc[subfield] = {}
+                out_doc = out_doc[subfield]
+            out_doc[parts[-1]] = out_value
+    return out_collection
+
+
 def _handle_out_stage(in_collection, database, options):
     # TODO(MetrodataTeam): should leave the origin collection unchanged
     out_collection = database.get_collection(options)
@@ -951,7 +1022,7 @@ def _handle_facet_stage(in_collection, database, options):
 
 
 _PIPELINE_HANDLERS = {
-    '$addFields': None,
+    '$addFields': _handle_add_fields_stage,
     '$bucket': _handle_bucket_stage,
     '$bucketAuto': None,
     '$collStats': None,
@@ -967,14 +1038,19 @@ _PIPELINE_HANDLERS = {
     '$listSessions': None,
     '$lookup': _handle_lookup_stage,
     '$match': lambda c, d, o: [doc for doc in c if filtering.filter_applies(o, doc)],
+    '$merge': None,
     '$out': _handle_out_stage,
+    '$planCacheStats': None,
     '$project': _handle_project_stage,
     '$redact': None,
     '$replaceRoot': _handle_replace_root_stage,
+    '$replaceWith': None,
     '$sample': _handle_sample_stage,
+    '$set': _handle_add_fields_stage,
     '$skip': lambda c, d, o: c[o:],
     '$sort': _handle_sort_stage,
     '$sortByCount': None,
+    '$unset': None,
     '$unwind': _handle_unwind_stage,
 }
 
