@@ -66,6 +66,11 @@ from mongomock.results import UpdateResult
 from mongomock.write_concern import WriteConcern
 from mongomock import WriteError
 
+if hasattr(time, 'perf_counter'):
+    _get_perf_counter = time.perf_counter
+else:
+    _get_perf_counter = time.clock
+
 _KwargOption = collections.namedtuple('KwargOption', ['typename', 'default', 'attrs'])
 
 _WITH_OPTIONS_KWARGS = {
@@ -77,7 +82,7 @@ _WITH_OPTIONS_KWARGS = {
         ('document', 'mode', 'mongos_mode', 'max_staleness')),
     'write_concern': _KwargOption(
         'pymongo.write_concern.WriteConcern', WriteConcern(),
-        ('acknowledged', 'document', 'is_server_default')),
+        ('acknowledged', 'document')),
     'read_concern': _KwargOption(
         'pymongo.read_concern.ReadConcern', None,
         ('document', 'level', 'ok_for_legacy'))
@@ -252,7 +257,7 @@ def _project_by_spec(doc, combined_projection_spec, is_include, container):
 
 
 class BulkOperationBuilder(object):
-    def __init__(self, collection, ordered=False):
+    def __init__(self, collection, ordered=False, bypass_document_validation=False):
         self.collection = collection
         self.ordered = ordered
         self.results = {}
@@ -260,13 +265,15 @@ class BulkOperationBuilder(object):
         self.done = False
         self._insert_returns_nModified = True
         self._update_returns_nModified = True
+        self._bypass_document_validation = bypass_document_validation
 
     def find(self, selector):
         return BulkWriteOperation(self, selector)
 
     def insert(self, doc):
         def exec_insert():
-            self.collection.insert_one(doc)
+            self.collection.insert_one(
+                doc, bypass_document_validation=self._bypass_document_validation)
             return {'nInserted': 1}
         self.executors.append(exec_insert)
 
@@ -362,12 +369,16 @@ class BulkOperationBuilder(object):
 
 class Collection(object):
 
-    def __init__(self, database, name, _db_store, write_concern=None, read_preference=None):
+    def __init__(
+            self, database, name, _db_store, write_concern=None, read_preference=None,
+            codec_options=None):
         self.database = database
         self._name = name
         self._db_store = _db_store
         self._write_concern = write_concern or WriteConcern()
         self._read_preference = read_preference or _READ_PREFERENCE_PRIMARY
+        self._codec_options = codec_options
+        self._read_concern = None
 
     def __repr__(self):
         return "Collection({0}, '{1}')".format(self.database, self.name)
@@ -404,13 +415,15 @@ class Collection(object):
             raise NotImplementedError(
                 'The codec options are not implemented in mongomock alone, you need to import '
                 'the pymongo library as well.')
-        return bson_codec_options.CodecOptions()
+        return self._codec_options
 
-    def initialize_unordered_bulk_op(self):
-        return BulkOperationBuilder(self, ordered=False)
+    def initialize_unordered_bulk_op(self, bypass_document_validation=False):
+        return BulkOperationBuilder(
+            self, ordered=False, bypass_document_validation=bypass_document_validation)
 
-    def initialize_ordered_bulk_op(self):
-        return BulkOperationBuilder(self, ordered=True)
+    def initialize_ordered_bulk_op(self, bypass_document_validation=False):
+        return BulkOperationBuilder(
+            self, ordered=True, bypass_document_validation=bypass_document_validation)
 
     def insert(self, data, manipulate=True, check_keys=True,
                continue_on_error=False, **kwargs):
@@ -419,16 +432,18 @@ class Collection(object):
         validate_write_concern_params(**kwargs)
         return self._insert(data)
 
-    def insert_one(self, document, session=None):
-        validate_is_mutable_mapping('document', document)
+    def insert_one(self, document, bypass_document_validation=False, session=None):
+        if not bypass_document_validation:
+            validate_is_mutable_mapping('document', document)
         return InsertOneResult(self._insert(document, session), acknowledged=True)
 
-    def insert_many(self, documents, ordered=True, session=None):
+    def insert_many(self, documents, ordered=True, bypass_document_validation=False, session=None):
         if not isinstance(documents, Iterable) or not documents:
             raise TypeError('documents must be a non-empty list')
         documents = list(documents)
-        for document in documents:
-            validate_is_mutable_mapping('document', document)
+        if not bypass_document_validation:
+            for document in documents:
+                validate_is_mutable_mapping('document', document)
         return InsertManyResult(
             self._insert(documents, session, ordered=ordered),
             acknowledged=True)
@@ -466,11 +481,6 @@ class Collection(object):
                 })
             return results
 
-        # Like pymongo, we should fill the _id in the inserted dict (odd behavior,
-        # but we need to stick to it), so we must patch in-place the data dict
-        for key in data.keys():
-            data[key] = helpers.patch_datetime_awareness_in_document(data[key])
-
         if not all(isinstance(k, string_types) for k in data):
             raise ValueError('Document keys must be strings')
 
@@ -478,14 +488,20 @@ class Collection(object):
             # bson validation
             BSON.encode(data, check_keys=True)
 
+        # Like pymongo, we should fill the _id in the inserted dict (odd behavior,
+        # but we need to stick to it), so we must patch in-place the data dict
         if '_id' not in data:
             data['_id'] = ObjectId()
+
         object_id = data['_id']
         if isinstance(object_id, dict):
             object_id = helpers.hashdict(object_id)
         if object_id in self._store:
             raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
-        self._store[object_id] = self._internalize_dict(data)
+
+        data = helpers.patch_datetime_awareness_in_document(data)
+
+        self._store[object_id] = data
         try:
             self._ensure_uniques(data)
         except DuplicateKeyError:
@@ -525,19 +541,26 @@ class Collection(object):
             sub_doc = sub_doc[part]
         return True
 
-    def update_one(self, filter, update, upsert=False, session=None):
-        validate_ok_for_update(update)
+    def update_one(
+            self, filter, update, upsert=False, bypass_document_validation=False, session=None):
+        if not bypass_document_validation:
+            validate_ok_for_update(update)
         return UpdateResult(self._update(filter, update, upsert=upsert, session=session),
                             acknowledged=True)
 
-    def update_many(self, filter, update, upsert=False, session=None):
-        validate_ok_for_update(update)
+    def update_many(
+            self, filter, update, upsert=False, bypass_document_validation=False, session=None):
+        if not bypass_document_validation:
+            validate_ok_for_update(update)
         return UpdateResult(self._update(filter, update, upsert=upsert,
                                          multi=True, session=session),
                             acknowledged=True)
 
-    def replace_one(self, filter, replacement, upsert=False, session=None):
-        validate_ok_for_replace(replacement)
+    def replace_one(
+            self, filter, replacement, upsert=False, bypass_document_validation=False,
+            session=None):
+        if not bypass_document_validation:
+            validate_ok_for_replace(replacement)
         return UpdateResult(self._update(filter, replacement, upsert=upsert, session=session),
                             acknowledged=True)
 
@@ -1334,10 +1357,9 @@ class Collection(object):
         if unknown_kwargs:
             raise OperationFailure("unrecognized field '%s'" % unknown_kwargs.pop())
 
-        count = len(list(self._iter_documents(filter)))
-        if limit is None:
-            return count - skip
-        return min(count - skip, limit)
+        doc_num = len(list(self._iter_documents(filter)))
+        count = max(doc_num - skip, 0)
+        return count if limit is None else min(count, limit)
 
     def estimated_document_count(self, **kwargs):
         if kwargs.pop('session', None):
@@ -1365,7 +1387,7 @@ class Collection(object):
         is_unique = kwargs.pop('unique', False)
         is_sparse = kwargs.pop('sparse', False)
 
-        index_string = helpers.gen_index_name(index_list)
+        index_name = kwargs.pop('name', helpers.gen_index_name(index_list))
         index_dict = {'key': index_list}
         if is_sparse:
             index_dict['sparse'] = True
@@ -1374,6 +1396,7 @@ class Collection(object):
         if is_unique:
             index_dict['unique'] = True
             indexed = set()
+            indexed_list = []
             for doc in self._store.documents:
                 index = []
                 for key, unused_order in index_list:
@@ -1386,13 +1409,19 @@ class Collection(object):
                 if is_sparse and not index:
                     continue
                 index = tuple(index)
-                if index in indexed:
-                    raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
-                indexed.add(index)
+                try:
+                    if index in indexed:
+                        raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
+                    indexed.add(index)
+                except TypeError:
+                    # index is not hashable.
+                    if index in indexed_list:
+                        raise DuplicateKeyError('E11000 Duplicate Key Error', 11000)
+                    indexed_list.append(index)
 
-        self._store.indexes[index_string] = index_dict
+        self._store.indexes[index_name] = index_dict
 
-        return index_string
+        return index_name
 
     def create_indexes(self, indexes, session=None):
         for index in indexes:
@@ -1404,7 +1433,8 @@ class Collection(object):
                 index.document['key'].items(),
                 session=session,
                 unique=index.document.get('unique', False),
-                sparse=index.document.get('sparse', False))
+                sparse=index.document.get('sparse', False),
+                name=index.document.get('name'))
             for index in indexes
         ]
 
@@ -1466,7 +1496,7 @@ class Collection(object):
             raise NotImplementedError('Mongomock does not handle sessions yet')
         if limit == 0:
             limit = None
-        start_time = time.clock()
+        start_time = _get_perf_counter()
         out_collection = None
         reduced_rows = None
         full_dict = {
@@ -1548,7 +1578,7 @@ class Collection(object):
             full_dict['result'] = reduced_rows
         else:
             raise TypeError("'out' must be an instance of string, dict or bson.SON")
-        full_dict['timeMillis'] = int(round((time.clock() - start_time) * 1000))
+        full_dict['timeMillis'] = int(round((_get_perf_counter() - start_time) * 1000))
         if full_response:
             ret_val = full_dict
         return ret_val
@@ -1629,22 +1659,29 @@ class Collection(object):
 
     def with_options(
             self, codec_options=None, read_preference=None, write_concern=None, read_concern=None):
+        has_changes = False
         for key, options in iteritems(_WITH_OPTIONS_KWARGS):
             value = locals()[key]
-            if value is None:
+            if value is None or value == getattr(self, '_' + key):
                 continue
+            has_changes = True
             for attr in options.attrs:
                 if not hasattr(value, attr):
                     raise TypeError(
                         '{} must be an instance of {}'.format(key, options.typename))
-            if key not in ('read_preference', 'write_concern') and options.default != value:
+            if key in ('codec_options', 'read_concern') and options.default != value:
+                # TODO(pascal): Support change of tz_aware in codec_options.
                 raise NotImplementedError(
                     '%s is a valid parameter for with_options but it is currently not implemented '
                     'in Mongomock' % key)
 
+        if not has_changes:
+            return self
+
         return Collection(
-            self.database, self.name, write_concern=write_concern, read_preference=read_preference,
-            _db_store=self._db_store)
+            self.database, self.name, write_concern=write_concern or self._write_concern,
+            read_preference=read_preference or self._read_preference,
+            codec_options=codec_options or self._codec_options, _db_store=self._db_store)
 
     def rename(self, new_name, session=None, **kwargs):
         if session:
@@ -1665,6 +1702,18 @@ class Collection(object):
             operation._add_to_bulk(bulk)
         return BulkWriteResult(bulk.execute(), True)
 
+    def find_raw_batches(
+            self, filter=None, projection=None, skip=0, limit=0, no_cursor_timeout=False,
+            cursor_type=None, sort=None, allow_partial_results=False,
+            oplog_replay=False, modifiers=None, batch_size=0, manipulate=True, collation=None,
+            hint=None, max_scan=None, max_time_ms=None, max=None, min=None, return_key=False,
+            how_record_id=False, snapshot=False, comment=None):
+        raise NotImplementedError('find_raw_batches method is not implemented in mongomock yet')
+
+    def aggregate_raw_batches(self, pipeline, **kwargs):
+        raise NotImplementedError(
+            'aggregate_raw_batches method is not implemented in mongomock yet')
+
 
 class Cursor(object):
 
@@ -1690,6 +1739,7 @@ class Cursor(object):
     def _compute_results(self, with_limit_and_skip=False):
         # Recompute the result only if the query has changed
         if not self._results or self._factory_last_generated_results != self._factory:
+            # TODO(pascal): Check collection's codec_options instead of reaching to the client.
             if self.collection.database.client._tz_aware:
                 results = [helpers.make_datetime_timezone_aware_in_document(x)
                            for x in self._factory()]
@@ -1737,6 +1787,9 @@ class Cursor(object):
         return self
 
     def count(self, with_limit_and_skip=False):
+        warnings.warn(
+            'count is deprecated. Use Collection.count_documents instead.',
+            DeprecationWarning, stacklevel=2)
         results = self._compute_results(with_limit_and_skip)
         return len(results)
 
