@@ -1,4 +1,5 @@
 from datetime import datetime
+import itertools
 
 from .helpers import ObjectId, RE_TYPE
 from . import OperationFailure
@@ -7,16 +8,35 @@ import numbers
 import operator
 import re
 from sentinels import NOTHING
-from six import iteritems, iterkeys, string_types
+from six import iteritems, iterkeys, string_types, PY3
 try:
     from types import NoneType
 except ImportError:
     NoneType = type(None)
 
-COMPILED_RE_TYPE = type(re.compile('a'))
+try:
+    from bson import Regex
+    _RE_TYPES = (RE_TYPE, Regex)
+except ImportError:
+    _RE_TYPES = (RE_TYPE,)
+
 _TOP_LEVEL_OPERATORS = {'$expr', '$text', '$where', '$jsonSchema'}
 
 
+_NOT_IMPLEMENTED_OPERATORS = {
+    '$bitsAllClear',
+    '$bitsAllSet',
+    '$bitsAnyClear',
+    '$bitsAnySet',
+    '$geoIntersects',
+    '$geoWithin',
+    '$maxDistance',
+    '$minDistance',
+    '$near',
+    '$nearSphere',
+}
+
+
 def filter_applies(search_filter, document):
     """Applies given filter
 
@@ -37,7 +57,7 @@ class _Filterer(object):
             '$in': _in_op,
             '$nin': lambda dv, sv: not _in_op(dv, sv),
             '$exists': lambda dv, sv: bool(sv) == (dv is not NOTHING),
-            '$regex': _not_nothing_and(lambda dv, sv: _regex(dv, re.compile(sv))),
+            '$regex': _not_nothing_and(_regex),
             '$elemMatch': self._elem_match_op,
             '$size': _size_op,
             '$type': _type_op
@@ -77,19 +97,34 @@ class _Filterer(object):
             if search == {'$exists': False} and not iter_key_candidates(key, document):
                 continue
 
+            if isinstance(search, dict) and '$all' in search:
+                if not self._all_op(iter_key_candidates(key, document), search['$all']):
+                    return False
+                continue
+
             for doc_val in iter_key_candidates(key, document):
                 has_candidates |= doc_val is not NOTHING
-                if isinstance(search, dict) and all(key.startswith('$') for key in search.keys()):
+                is_ops_filter = search and isinstance(search, dict) and \
+                    all(key.startswith('$') for key in search.keys())
+                if is_ops_filter:
                     if '$options' in search and '$regex' in search:
                         search = _combine_regex_options(search)
-                    is_match = (all(
+                    unknown_operators = set(search) - set(self._operator_map) - {'$not'}
+                    if unknown_operators:
+                        not_implemented_operators = unknown_operators & _NOT_IMPLEMENTED_OPERATORS
+                        if not_implemented_operators:
+                            raise NotImplementedError(
+                                "'%s' is a valid operation but it is not supported by Mongomock "
+                                'yet.' % list(not_implemented_operators)[0])
+                        raise OperationFailure('unknown operator: ' + list(unknown_operators)[0])
+                    is_match = all(
                         operator_string in self._operator_map and
                         self._operator_map[operator_string](doc_val, search_val) or
                         operator_string == '$not' and
                         self._not_op(document, key, search_val)
                         for operator_string, search_val in iteritems(search)
-                    ) and search) or doc_val == search
-                elif isinstance(search, RE_TYPE) and isinstance(doc_val, (string_types, list)):
+                    ) and search
+                elif isinstance(search, _RE_TYPES) and isinstance(doc_val, (string_types, list)):
                     is_match = _regex(doc_val, search)
                 elif key in LOGICAL_OPERATOR_MAP:
                     if not search:
@@ -120,7 +155,7 @@ class _Filterer(object):
             for key in s.keys():
                 if key not in self._operator_map and key not in LOGICAL_OPERATOR_MAP:
                     raise OperationFailure('unknown operator: %s' % key)
-        elif isinstance(s, type(re.compile(''))):
+        elif isinstance(s, _RE_TYPES):
             pass
         else:
             raise OperationFailure('$not needs a regex or a document')
@@ -134,6 +169,8 @@ class _Filterer(object):
         return any(self.apply(query, item) for item in doc_val)
 
     def _all_op(self, doc_val, search_val):
+        if isinstance(doc_val, list) and doc_val and isinstance(doc_val[0], list):
+            doc_val = list(itertools.chain.from_iterable(doc_val))
         dv = _force_list(doc_val)
         matches = []
         for x in search_val:
@@ -205,10 +242,12 @@ def _force_list(v):
 
 
 def _in_op(doc_val, search_val):
+    if not isinstance(search_val, (list, tuple)):
+        raise OperationFailure('$in needs an array')
     if doc_val is NOTHING and None in search_val:
         return True
     doc_val = _force_list(doc_val)
-    is_regex_list = [isinstance(x, COMPILED_RE_TYPE) for x in search_val]
+    is_regex_list = [isinstance(x, _RE_TYPES) for x in search_val]
     if not any(is_regex_list):
         return any(x in search_val for x in doc_val)
     for x, is_regex in zip(search_val, is_regex_list):
@@ -267,6 +306,14 @@ def bson_compare(op, a, b, can_compare_types=True):
     if isinstance(a, NoneType):
         return op(0, 0)
 
+    # bson handles bytes as binary in python3+:
+    # https://api.mongodb.com/python/current/api/bson/index.html
+    if PY3 and isinstance(a, bytes):
+        # Performs the same operation as described by:
+        # https://docs.mongodb.com/manual/reference/bson-type-comparison-order/#bindata
+        if len(a) != len(b):
+            return op(len(a), len(b))
+        # bytes is always treated as subtype 0 by the bson library
     return op(a, b)
 
 
@@ -289,11 +336,14 @@ def _get_compare_type(val):
         return 20
     if isinstance(val, (tuple, list)):
         return 25
+    if isinstance(val, bytes):
+        assert PY3
+        return 30
     if isinstance(val, ObjectId):
         return 35
     if isinstance(val, datetime):
         return 45
-    if isinstance(val, RE_TYPE):
+    if isinstance(val, _RE_TYPES):
         return 50
     raise NotImplementedError(
         "Mongomock does not know how to sort '%s' of type '%s'" %
@@ -303,6 +353,11 @@ def _get_compare_type(val):
 def _regex(doc_val, regex):
     if not (isinstance(doc_val, (string_types, list)) or isinstance(doc_val, RE_TYPE)):
         return False
+    if isinstance(regex, string_types):
+        regex = re.compile(regex)
+    if not isinstance(regex, RE_TYPE):
+        # bson.Regex
+        regex = regex.try_compile()
     return any(
         regex.search(item) for item in _force_list(doc_val)
         if isinstance(item, string_types))
@@ -316,7 +371,7 @@ def _size_op(doc_val, search_val):
 
 def _list_expand(f, negative=False):
     def func(doc_val, search_val):
-        if isinstance(doc_val, (list, tuple)):
+        if isinstance(doc_val, (list, tuple)) and not isinstance(search_val, (list, tuple)):
             if negative:
                 return all(f(val, search_val) for val in doc_val)
             return any(f(val, search_val) for val in doc_val)
@@ -352,14 +407,19 @@ def _combine_regex_options(search):
     if options is None:
         return search_copy
 
-    if isinstance(search['$regex'], COMPILED_RE_TYPE):
+    if isinstance(search['$regex'], _RE_TYPES):
         keys = [k for k in iterkeys(search) if k in {'$regex', '$options'}]
         if keys == ['$options', '$regex']:
             raise NotImplementedError(
                 'Do not use compiled regular expressions with $options until '
                 'https://jira.mongodb.org/browse/SERVER-38621 is solved.')
-        search_copy['$regex'] = re.compile(
-            search['$regex'].pattern, search['$regex'].flags | options)
+        if isinstance(search['$regex'], RE_TYPE):
+            search_copy['$regex'] = re.compile(
+                search['$regex'].pattern, search['$regex'].flags | options)
+        else:
+            # bson.Regex
+            regex = search['$regex']
+            search_copy['$regex'] = regex.__class__(regex.pattern, regex.flags | options)
     else:
         search_copy['$regex'] = re.compile(search['$regex'], options)
     return search_copy
